#include <QCoreApplication>
#include <cassert>
#include <cmath>
#include <iostream>
#include <string>
#include <unordered_set>
#include <QJsonArray>
#include <QJsonObject>
#include <QJsonDocument>
#include <QIODevice>
#include <QFile>
#include <QDir>
#include <QTimer>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QNetworkRequest>

using namespace std;
unordered_map<string, vector<double>> tradePrices;
unordered_map<string, int> tradeStocks;
QString timestamp;

// Multipliers for each currency, normalized to emeralds
unordered_map<string, double> currencyMultipliers;
vector<string> validCurrencies;

QString dataURL = "http://pvc-utils.xyz/pvc-prices-history/data.json";
QString currencyMultipliersURL = "http://pvc-utils.xyz/resources/currencymultipliers.json";


bool fileExists(const QString& filename) {
    QFile file(filename);
    return file.exists();
}

double roundToSignificantDigits(double value, int digits) {
    if (value == 0.0) // otherwise it will return 'nan' due to the log10() of zero
        return 0.0;

    double factor = pow(10.0, digits - ceil(log10(fabs(value))));
    return round(value * factor) / factor;
}

double roundToDecimalPlace(double num, int decimalPlaces) {
    double scale = pow(10.0, decimalPlaces);
    return round(num * scale) / scale;
}

QJsonDocument downloadJson(QUrl URL){
    QNetworkAccessManager* manager = new QNetworkAccessManager();


    QNetworkRequest networkRequest(URL);
    QNetworkReply* reply = manager->get(networkRequest);

    QEventLoop eventLoop;
    QObject::connect(reply, &QNetworkReply::finished, &eventLoop, &QEventLoop::quit);
    eventLoop.exec();

    QByteArray replyData = reply->readAll();
    QJsonDocument jsonDocument = QJsonDocument::fromJson(replyData);
    reply->deleteLater();
    return jsonDocument;
}



double getMedian(const vector<double>& data) {
    if (data.empty()) {
        return 0.0; // Handle the case of an empty vector
    }

    vector<double> sortedData = data; // Create a copy of the data
    sort(sortedData.begin(), sortedData.end());

    size_t size = sortedData.size();
    if (size % 2 == 0) {
        // If the size is even, return the average of the two middle values
        size_t mid = size / 2;
        return (sortedData[mid - 1] + sortedData[mid]) / 2.0;
    } else {
        // If the size is odd, return the middle value
        return sortedData[size / 2];
    }
}

void saveTradePrices(const unordered_map<string, vector<double>>& tradePrices, const unordered_map<string, int>& tradeStocks, const QString& timestamp) {
    QByteArray jsonData;
    QJsonObject rootObject;
    rootObject["timeDumped"] = timestamp;

    QJsonObject dataObject;
    for (const auto &entry : tradePrices) {
        double median = roundToDecimalPlace(roundToSignificantDigits(getMedian(entry.second), 7), 6);

        // Create a QJsonArray to represent the tuple
        QJsonArray tuple;
        tuple.append(median); // Add the price
        tuple.append(tradeStocks.at(entry.first)); // Add the stock

        dataObject[QString::fromStdString(entry.first)] = tuple;
    }

    rootObject["data"] = dataObject;

    QJsonDocument saveDoc(rootObject);
    jsonData = saveDoc.toJson(QJsonDocument::Compact);

    // Create a hash of the file content
    QByteArray hash = QCryptographicHash::hash(jsonData, QCryptographicHash::Md5);

    // Save the file with the hash as its name
    QFile file(QString::fromStdString("data/") + hash.toHex() + ".json");
    if (!file.open(QIODevice::WriteOnly)) {
        // Handle error opening file
        cout << "Failed opening file!" << endl;
        return;
    }

    file.write(jsonData);
    cout << "Data size: " << jsonData.size() << endl << "File hash: " << hash.toHex().toStdString() << endl;
    file.close();
}

void deleteOldFiles(const QString& directoryPath) {
    QDir dir(directoryPath);
    QStringList fileList = dir.entryList(QDir::Files);

    QDate currentDate = QDate::currentDate();
    QDate thirtyDaysAgo = currentDate.addDays(-30);

    for (const QString& file : fileList) {
        QFileInfo fileInfo(dir.filePath(file));
        QDate lastModifiedDate = fileInfo.lastModified().date();

        if (lastModifiedDate < thirtyDaysAgo) {
            if (!QFile::remove(dir.filePath(file))) {
                // Handle error deleting file
                continue;
            }
        }
    }
}

void storeFileList(const QString& directoryPath) {
    QDir dir(directoryPath);
    QStringList fileList = dir.entryList(QDir::Files);

    QJsonArray jsonArray;
    for (const QString& file : fileList) {
        jsonArray.append(file);
    }

    QJsonDocument doc(jsonArray);
    QFile file("filelist.json");
    if (!file.open(QIODevice::WriteOnly)) {
        // Handle error opening file
        return;
    }

    file.write(doc.toJson());
    file.close();
}

// Normalize prices to emeralds
double normalizePrices(const string& currency, double price, unordered_map<string, double> currencyMultipliers) {

    auto it = currencyMultipliers.find(currency);
    if (it != currencyMultipliers.end()) {
        double multiplier = it->second;
        price *= multiplier;
    } else {
        price = 0;
    }

    return price;
}

void parse(const QJsonObject& data, unordered_map<string, vector<double>>& tradePrices, unordered_map<string, int>& tradeStocks, unordered_map<string, double> currencyMultipliers, vector<string> validCurrencies) {
    cout << "Parsing JSON data..." << endl;
    tradePrices.clear();

    int ignoredDuplicates = 0;
    int ignoredProbableDonations = 0;

    unordered_set<string> validCurrenciesSet(validCurrencies.begin(), validCurrencies.end());

    if (data.contains("data") && data["data"].isArray()) {
        QJsonArray dataArray = data["data"].toArray();

        for (const QJsonValue& entryValue : dataArray) {
            unordered_set<string> currentShop;
            unordered_set<int> currentShopStocks;
            unordered_set<string> currentShopEnchants;

            QJsonObject entry = entryValue.toObject();
            // For each trade fetch all necessary values
            if (entry.contains("recipes") && entry["recipes"].isArray()) {
                QJsonArray recipesArray = entry["recipes"].toArray();
                for (const QJsonValue& recipeValue : recipesArray) {
                    QJsonObject recipe = recipeValue.toObject();

                    QJsonObject item1 = recipe["item1"].toObject();
                    QString currency = item1["type"].toString();
                    double price = item1["amount"].toDouble();
                    QString currency2;
                    double price2 = 0.0;

                    if (recipe.contains("item2")) {
                        QJsonObject item2 = recipe["item2"].toObject();
                        currency2 = item2["type"].toString();
                        price2 = item2["amount"].toDouble();
                    }

                    QJsonObject resultItem = recipe["resultItem"].toObject();
                    QString trade = resultItem["type"].toString();
                    int tradeAmount = resultItem["amount"].toInt();

                    QJsonObject enchantObject = resultItem["enchant"].toObject();
                    string enchants = "Default Value";
                    for (auto it = enchantObject.begin(); it != enchantObject.end(); ++it) {
                        enchants += it.key().toStdString();
                    }

                    int stock = recipe["stock"].toInt() * tradeAmount;
                    bool validity = true; // Flag to track if the trade is valid

                    // Handle shulker boxes
                    // Adds all items inside the shulker box and sets the trade item
                    // To shulker box contents
                    // Ignores shulkers with several item types inside

                    if (trade.contains("SHULKER_BOX")) {
                        QJsonArray loreEntries = recipe["resultItem"].toObject()["lore"].toArray();
                        if (!loreEntries.isEmpty()) {
                            // Check if all lore entries are identical
                            bool identical = adjacent_find(loreEntries.begin(), loreEntries.end(), not_equal_to<>()) == loreEntries.end();
                            if (identical) {

                                // Reinitialize stock and trade
                                stock = 0;
                                trade = "";
                                tradeAmount = 0;

                                // Extract the name and amount from the lore entry
                                QString loreEntry = loreEntries[0].toString().mid(2); // Remove the leading '- '
                                QStringList parts = loreEntry.split(" ", Qt::SkipEmptyParts);

                                if (parts.size() >= 2) {
                                    try {
                                        int lore_entry_multiplier = parts[0].mid(0, parts[0].length() - 1).toInt(); // Extract the lore entry multiplier by removing the 'x' character
                                        QString lore_text = parts[1];

                                        // Append item name in the lore entry to trade
                                        tradeAmount += lore_entry_multiplier * loreEntries.size();
                                        stock = lore_entry_multiplier * loreEntries.size();
                                        trade = lore_text;
                                    } catch (const invalid_argument& e) {
                                        cerr << "Invalid lore entry format: " << loreEntry.toStdString() << endl;
                                        validity = false; // Set the flag to indicate an invalid trade
                                    }
                                } else {
                                    validity = false;
                                }
                                stock *= recipe["stock"].toInt();
                            } else {
                                validity = false;
                            }
                        }
                    }

                    // Handle invalid currencies
                    if (validCurrenciesSet.find(currency.toStdString()) == validCurrenciesSet.end()) {
                        price = 0.0;
                    }
                    if (!currency2.isEmpty() && validCurrenciesSet.find(currency2.toStdString()) == validCurrenciesSet.end()) {
                        price2 = 0.0;
                    }

                    price = normalizePrices(currency.toStdString(), price, currencyMultipliers) + normalizePrices(currency2.toStdString(), price2, currencyMultipliers);
                    if (price == 0.0) {
                        validity = false;
                    }

                    double pricePerUnit = price / tradeAmount;

                    // Create the trade instance and add price to prices dictionary
                    if (validity) {
                        if (tradePrices.find(trade.toStdString()) == tradePrices.end()) {
                            tradePrices[trade.toStdString()] = vector<double>();
                        }

                        if (stock != 0) {
                            tradePrices[trade.toStdString()].push_back(pricePerUnit);
                            if (tradeStocks.find(trade.toStdString()) == tradeStocks.end()) {
                                tradeStocks[trade.toStdString()] = 0;
                            }

                            bool insertedTrade = currentShop.insert(trade.toStdString()).second;
                            bool insertedStock = currentShopStocks.insert(stock).second;
                            bool insertedEnchants = currentShopEnchants.insert(enchants).second;

                            if(insertedTrade || insertedStock || insertedEnchants) {
                                tradeStocks[trade.toStdString()] += stock;
                            } else {
                                ignoredDuplicates++;
                            }
                        }
                    }
                }

            }
        }
    }

    int price_threshold = 10;
    // For each item in the tradePrices dictionary
    for (auto it = tradePrices.begin(); it != tradePrices.end(); ) {

        vector<double>& prices = it->second;

        // Calculate the median price
        double medianPrice = getMedian(prices);

        // Iterate over prices
        for (auto i = 0; i < prices.size(); ) {
            if (prices[i] > price_threshold * medianPrice) {
                // Erase the price from the vector
                prices.erase(prices.begin() + i);
                ignoredProbableDonations++;

                // After erasing, check if i is now beyond the last element
                if (i >= prices.size()) {
                    break; // Exit the loop if no more elements are left to check
                }
            } else {
                // Only increment i if no erasure has happened
                ++i;
            }
        }
        // If all prices were removed, remove the item from the map
        if (prices.empty()) {
            it = tradePrices.erase(it);
        } else {
            ++it;
        }
    }
    cout << "Parsed JSON. Removed " << ignoredProbableDonations << " probable donation shops and " << ignoredDuplicates << " probable duplicates from stock counting. There are a total of " << tradePrices.size() << " item types on sale." << endl;
}

int main(int argc, char* argv[]) {
    QCoreApplication app(argc, argv);

    QTimer::singleShot(0, [](){
        // Fetch currency multipliers
        QJsonObject currencyMultipliersJson = downloadJson(currencyMultipliersURL).object();
        std::unordered_map<std::string, double> currencyMultipliers;

        for (auto it = currencyMultipliersJson.begin(); it != currencyMultipliersJson.end(); ++it) {
            QString key = it.key();
            QJsonValue value = it.value();

            // Ensure that the value is a double before inserting
            if (value.isDouble()) {
                currencyMultipliers.insert({key.toStdString(), value.toDouble()});
                validCurrencies.emplace_back(key.toStdString());
            }
        }

        auto dataJson = downloadJson(dataURL).object();
        assert (! dataJson.isEmpty());
        parse(dataJson, tradePrices, tradeStocks, currencyMultipliers, validCurrencies);

        saveTradePrices(tradePrices, tradeStocks, dataJson.value("information").toObject().value("timeDumped").toString());
        storeFileList("data/");
        deleteOldFiles("data/");
        QCoreApplication::exit(0);
    });
    return app.exec();
}
